#! python3  # noqa: E265

"""
    Localization module.

    Author: Isogeo

    See: https://docs.python.org/fr/3/library/gettext.html#module-gettext
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import gettext
import locale
import logging
from pathlib import Path

# package
from scan_metadata_processor.__about__ import __package_name__

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)


# #############################################################################
# ########## Functions #############
# ##################################


class Localizer(object):
    """Localization module.

    :param str lang_code: language code to use for translation. Defaults to: None - optional
    """

    LANGUAGE_CODE_DEFAULT = "en_GB"
    LANGUAGE_CODE_TO_USE = "fr_FR"

    tr = gettext.gettext

    def __init__(self, lang_code: str = "fr_FR"):
        """Instanciation method."""
        if lang_code is not None:
            self.LANGUAGE_CODE_TO_USE = lang_code
            self.locale_setter(lang_code)
        else:
            # if not set, use system
            self.LANGUAGE_CODE_TO_USE = locale.getdefaultlocale()[0]
            self.locale_setter(locale.getdefaultlocale()[0])

        # init translator with required language
        self.init_translator()

    def init_translator(self) -> gettext.gettext:
        """Handy shortcut to initialize the translator functions.

        :return: gettext translator method
        :rtype: gettext.gettext

        :example:

        .. code-block:: python

            _ = init_translator()
            print(_("Hello Mister Computer!"))
        """
        localedir = Path(__file__).parent / Path("../i18n")
        gettext.bindtextdomain(__package_name__, localedir)
        try:
            translate = gettext.translation(
                domain=__package_name__,
                localedir=str(localedir.resolve()),
                fallback=False,
                languages=[self.LANGUAGE_CODE_TO_USE],
            )
        except FileNotFoundError:
            if not self.LANGUAGE_CODE_TO_USE.startswith("en"):
                logger.warning(
                    "No translation found for language: {}. Using raw texts.".format(
                        self.LANGUAGE_CODE_TO_USE
                    )
                )
            self.LANGUAGE_CODE_TO_USE = self.LANGUAGE_CODE_DEFAULT
            translate = gettext.translation(
                domain=__package_name__,
                localedir=str(localedir.resolve()),
                fallback=True,
                languages=[self.LANGUAGE_CODE_TO_USE],
            )

        translate.install(__package_name__)
        logger.debug(
            "Translation loaded for language: {}".format(self.LANGUAGE_CODE_TO_USE)
        )
        self.tr = translate.gettext
        return translate.gettext

    @staticmethod
    def locale_setter(expected_locale: str = "en_EN"):
        """Ensure locale is the expected one.

        :param str expected_locale: locale code to set. Defaults to: "fr_FR" - optional
        """
        # get the default locale on system
        default_locale = locale.getdefaultlocale()  # -> tuple

        # compare with the expected locale
        if default_locale[0] == expected_locale:
            logger.debug(
                "Default locale is matching the expected one: {}".format(
                    expected_locale
                )
            )
            locale.setlocale(locale.LC_ALL, "")
        else:
            logger.debug(
                "Default locale ({}) is not the expected one: {}. "
                "Change will be effective until the end of program.".format(
                    default_locale, expected_locale
                )
            )
            try:
                locale.setlocale(locale.LC_ALL, expected_locale)
            except locale.Error as err:
                logger.warning(
                    "Expected locale is not available on this system or "
                    "application is not translated intot this. "
                    "Falling back to default locale: {}. Trace: {}.".format(
                        default_locale[0], err
                    )
                )
                locale.resetlocale()



#! python3  # noqa: E265

"""
    Shortcuts to exit CLI properly and with pretty messages.

    Author: Isogeo
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging

# 3rd party
import click

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# #############################################################################
# ########## Functions #############
# ##################################


def exit_cli_error(message: str, abort: bool = True):
    """Display error message (red) and stop execution.

    :param str message: message to log and display in terminal.
    :param bool abort: option to abort after displaying . Defaults to: True - optional
    """
    # log
    logger.error(message)

    # handle cases when the full exception is passed
    if isinstance(message, Exception):
        if isinstance(message.args, tuple) and len(message.args):
            message = message.args[0]
        else:
            message = getattr(message, "message", repr(message))

    # echo to terminal
    click.secho(message=message, err=True, fg="red")
    if abort:
        click.Context.abort(message)


def exit_cli_normal(message: str, abort: bool = True):
    """Display normal message (magenta) and stop execution.

    :param str message: message to log and display in terminal.
    :param bool abort: option to abort after displaying . Defaults to: True - optional
    """
    logger.info(message)
    click.secho(message=message, err=False, fg="magenta")

    if abort:
        click.Context.abort(message)


def exit_cli_success(message: str, abort: bool = True):
    """Display success message (green) and stop execution.

    :param str message: message to log and display in terminal.
    :param bool abort: option to abort after displaying the message. Defaults to: True - optional
    """
    logger.info(message)
    click.secho(message=message, err=False, fg="green")
    if abort:
        click.Context.abort(message)


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265

"""
    Synchroniser between loaded metadata from Scan and Isogeo database through API.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from concurrent.futures import ThreadPoolExecutor
from time import gmtime, sleep, strftime

# 3rd party
from isogeo_pysdk import Event, FeatureAttribute, Isogeo, Metadata, Workgroup
from scan_metadata_processor.comparator import MetadataComparator
from scan_metadata_processor.converter import LookupToIsogeo
from scan_metadata_processor.converter.matchers import match_coordinate_system

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)

# ##############################################################################
# ########## Classes ###############
# ##################################


class MetadataSynchronizer:
    """Load metadata produced by the Scan and send it to Isogeo.

    :param Isogeo api_client: authenticated API client
    :param object report: report object in which to store process results
    """

    def __init__(self, api_client: Isogeo, report: object = None):
        """Instanciating metadata synchronizer."""
        # Isogeo API client
        self.isogeo = api_client
        self.isogeo_formats_codes = [
            i.get("code") for i in self.isogeo.formats.listing()
        ]

        # report object to fill
        self.report = report

    def apply_coordinate_system(
        self, target_metadata: Metadata, srs_matcher_result: tuple, group: Workgroup
    ):
        """Add a coordinate-system to a metadata according to the SRS matcher result.

        :param Metadata target_metadata: metadata to which to add the coordinate-system
        :param tuple srs_matcher_result: result of the SRS matcher
        :param Workgroup group: group to which to add the missing coordinate-system

        :raises ValueError: for unexpected srs matcher result
        """
        if srs_matcher_result[0] == 0:
            logger.warning(
                "No coordinate-system identified for: {}".format(
                    target_metadata.title_or_name(slugged=True)
                )
            )
            self.report.li_srs_not_identified.append(target_metadata._id)
        elif srs_matcher_result[0] == 1:
            self.isogeo.srs.associate_metadata(
                metadata=target_metadata, coordinate_system=srs_matcher_result[1]
            )
        elif srs_matcher_result[0] == 2:
            self.isogeo.srs.associate_workgroup(
                coordinate_system=srs_matcher_result[1], workgroup=group
            )
            self.isogeo.srs.associate_metadata(
                metadata=target_metadata, coordinate_system=srs_matcher_result[1]
            )
        elif srs_matcher_result[0] == 3:
            logger.warning(
                "Coordinate-system not recognized by FME for: {}".format(
                    target_metadata.title_or_name(slugged=True)
                )
            )
        else:
            raise ValueError(
                "Unexpected value returned by coordinate-system matcher: {}".format(
                    srs_matcher_result
                )
            )

    def create_from_lookup(
        self,
        group: Workgroup,
        group_catalogs: list,
        group_coordsys: list,
        converter: LookupToIsogeo,
    ) -> Metadata:
        """Create a metadata from a Scan lookup model.

        :param Workgroup group: Isogeo group in which to create the metadata
        :param list group_catalogs: list of the group's catalogs with the option "add to scan"
        :param list group_coordsys: list of group selected coordinate-systems
        :param LookupToIsogeo converter: lookup converter to Isogeo metadata

        :return: created metadata
        :rtype: Metadata
        """
        # create metadata - handle hard (code) and soft (API) errors
        try:
            md_created = self.isogeo.metadata.create(
                workgroup_id=group._id, metadata=converter.as_metadata()
            )
        except Exception as err:
            err_msg = "Failed to create a metadata: {}. Trace: {}".format(
                converter.extract_name(), err
            )
            logger.error(err_msg)
            self.report.counter_errorred += 1
            return

        # check if API could create the metadata
        if not isinstance(md_created, Metadata):
            if converter.in_lookup.envelope is not None:
                logger.warning(
                    "Creation failed. Try again without the geographic envelope "
                    "because of left-hand rule handling in Isogeo API."
                )
                converter.in_lookup.envelope = None
                self.report.li_envelope_incorrect.append(converter.extract_name())
                return self.create_from_lookup(
                    group, group_catalogs, group_coordsys, converter
                )
            logger.error(
                "Metadata creation failed for '{}'. Error: {}".format(
                    converter.extract_name(), md_created
                )
            )
            self.report.counter_errorred += 1
            return
        else:
            self.report.counter_created += 1

        # have a break, have a kitkat (API needs a breathe...)
        sleep(0.5)
        logger.debug(
            "MD created: {} ({})".format(
                md_created._id, md_created.title_or_name(slugged=True)
            )
        )

        # creation event
        try:
            md_evt_creation = converter.prepare_event(evt_type="creation")
            self.isogeo.metadata.events.create(
                metadata=md_created, event=md_evt_creation
            )
        except Exception as err:
            err_msg = "Failed to add an event to metadata: {} ({}). Trace: {}".format(
                md_created.title_or_name(slugged=True), md_created._id, err
            )
            logger.error(err_msg)

        # SRS
        try:
            srs_match = match_coordinate_system(
                group_srs=group_coordsys, md_lookup=converter.in_lookup
            )
            self.apply_coordinate_system(
                target_metadata=md_created, srs_matcher_result=srs_match, group=group
            )
        except Exception as err:
            err_msg = "Failed to get and associate a coordinate-system to metadata: {} ({}). Trace: {}".format(
                md_created.title_or_name(slugged=True), md_created._id, err
            )
            logger.error(err_msg)

        # feature attributes
        try:
            with ThreadPoolExecutor(
                thread_name_prefix="AddFeatureAttributes"
            ) as executor:
                for featattr in converter.prepare_feature_attributes():
                    executor.submit(
                        self.isogeo.metadata.attributes.create,
                        metadata=md_created,
                        attribute=featattr,
                    )
        except Exception as err:
            err_msg = "Failed to add feature attributes to metadata: {} ({}). Trace: {}".format(
                md_created.title_or_name(slugged=True), md_created._id, err
            )
            logger.error(err_msg)

        # catalogs
        if isinstance(group_catalogs, list) and len(group_catalogs):
            try:
                self.isogeo.metadata.bulk.prepare(
                    metadatas=(md_created._id,),
                    action="add",
                    target="catalogs",
                    models=tuple(group_catalogs),
                )
                self.isogeo.metadata.bulk.send()
            except Exception as err:
                err_msg = "Failed to associate all catalogs to metadata: {} ({}). Trace: {}".format(
                    md_created.title_or_name(slugged=True), md_created._id, err
                )
                logger.warning(err_msg)
                self.isogeo.catalog.associate_metadata(
                    metadata=md_created, catalog=group_catalogs[0]
                )
        else:
            logger.warning("No catalogs to associate")

        # return signature
        return md_created

    def update_from_comparator(self, comparator: MetadataComparator) -> Metadata:

        """Create a metadata from a Scan lookup model.

        :param MetadataComparator comparator: comparator object used to perform \
        deletion/addition operations

        :return: created metadata
        :rtype: Metadata
        """
        # extract objects from comparator
        md_to_update = comparator.md_old
        md_from_lookup = comparator.md_new
        evt_diff = Event(
            date=strftime("%Y-%m-%d", gmtime()),
            description=comparator.diff_message(),
            kind="update",
        )

        # update features count
        if comparator.diff_features_count():
            logger.debug("Updating features count for: {}".format(md_to_update._id))
            md_to_update.features = md_from_lookup.features
            self.isogeo.metadata.update(metadata=md_to_update)

        # update coordinate-system
        if comparator.diff_srs():
            logger.debug("Changing coordinate-system for: {}".format(md_to_update._id))
            self.isogeo.coordinate_system.associate_metadata(
                metadata=md_to_update, coordinate_system=md_from_lookup.coordinateSystem
            )

        # update feature attributes
        diff_attrs = comparator.diff_features_attributes()
        if diff_attrs[0]:
            # add new attributes
            try:
                with ThreadPoolExecutor(
                    thread_name_prefix="AddFeatureAttributes"
                ) as executor:
                    for new_attr in diff_attrs[1]:
                        executor.submit(
                            self.isogeo.metadata.attributes.create,
                            metadata=md_to_update,
                            attribute=FeatureAttribute(
                                name=new_attr[0], dataType=new_attr[1]
                            ),
                        )
            except Exception as err:
                err_msg = "Failed to add feature attributes to metadata: {} ({}). Trace: {}".format(
                    md_to_update.title_or_name(slugged=True), md_to_update._id, err
                )
                logger.error(err_msg)

            # remove deleted attributes
            # we need the attribute _id, so make a match
            match_attr_to_remove = [
                attr
                for attr in md_to_update.featureAttributes
                if attr.name in diff_attrs[2]
            ]
            try:
                with ThreadPoolExecutor(
                    thread_name_prefix="RemoveFeatureAttributes"
                ) as executor:
                    for attr in match_attr_to_remove:
                        executor.submit(
                            self.isogeo.metadata.attributes.delete,
                            attribute=attr,
                            metadata=md_to_update,
                        )
            except Exception as err:
                err_msg = "Failed to add feature attributes to metadata: {} ({}). Trace: {}".format(
                    md_to_update.title_or_name(slugged=True), md_to_update._id, err
                )
                logger.error(err_msg)

        # add update event
        try:
            self.isogeo.metadata.events.create(metadata=md_to_update, event=evt_diff)
        except Exception as err:
            logger.error(
                "Failed to add update event to metadata: {}. Trace: {}".format(
                    md_to_update.admin_url(), err
                )
            )


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    pass



#! python3  # noqa: E265

"""
    Matching table between FME and Isogeo geometry names.

    Copied from: https://github.com/isogeo/isogeo-worker-server/blob/520791363f3a8216947e573851e9e22829a11446/lib/sync/utils.js#L9-L20
"""

MATCHER_GEOMETRY = {
    "fme_point": "Point",
    "fme_area": "Polygon",
    "fme_line": "LineString",
    "fme_arc": "Curve",
    "fme_collection": "GeometryCollection",
    "fme_ellipse": "Curve",
    "fme_surface": "Surface",
    "fme_point_cloud": "Point",
}


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    pass



#! python3  # noqa E265

"""
    Isogeo Scan - Model of metadata output of Sign vector.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# standard library
import logging
import pprint
from typing import Union


# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################
class Sign(object):
    """Output from FME scripts sign-*."""

    ATTR_TYPES = {
        "error": list,
        "signatures": Union[str, list],
    }

    def __init__(
        self,
        feature: dict = None,
        error: list = None,
        signatures: Union[str, list] = None,
        **kwargs,
    ):
        # default values for the object error/properties
        self._feature = None
        self._error = None
        self._signatures = None

        # if values have been passed, so use them as objects error.
        # error are prefixed by an underscore '_'
        if feature is not None:
            # case when the JSON is passed
            error = feature.get("error")
            signatures = feature.get("signatures")
        if error is not None:
            self._error = error
        if signatures is not None:
            if isinstance(signatures, str):
                signatures = [signatures]
            self._signatures = signatures

        # warn about unsupported attributes
        if len(kwargs):
            logger.warning(
                "Folllowings fields were not expected and have been ignored. "
                "Maybe consider adding them to the model: {}.".format(
                    " | ".join(kwargs.keys())
                )
            )

    # -- PROPERTIES --------------------------------------------------------------------
    # error
    @property
    def error(self) -> str:
        """Gets the error message.

        :return: error message recorded by the Scan.
        :rtype: str
        """
        return self._error

    @error.setter
    def error(self, error: str):
        """Sets the error message.

        :param str error: error message recorded by the Scan.
        """

        self._error = error

    # signatures
    @property
    def signatures(self) -> list:
        """Gets the signatures of this Metadata.

        :return: The signatures of this Metadata.
        :rtype: list
        """
        return self._signatures

    @signatures.setter
    def signatures(self, signatures: list):
        """Sets the signatures of this Metadata.

        :param list signatures: to be set
        """
        self._signatures = signatures

    # -- METHODS -----------------------------------------------------------------------
    def to_dict(self) -> dict:
        """Returns the model properties as a dict."""
        result = {}

        for attr, _ in self.ATTR_TYPES.items():
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value
        if issubclass(Sign, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model."""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other) -> bool:
        """Returns true if both objects are equal."""
        if not isinstance(other, Sign):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        """Returns true if both objects are not equal."""
        return not self == other


# ##############################################################################
# ##### Stand alone program ########
# ##################################
if __name__ == "__main__":
    """standalone execution."""
    sample = Sign()



#! python3  # noqa: E265


"""
    Sub-command in charge of checking settings and environment.

    Author: Isogeo
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os import R_OK, W_OK, access, environ, getlogin
from pathlib import Path
from timeit import default_timer

# 3rd party library
import click
from isogeo_pysdk.checker import IsogeoChecker
from requests import HTTPError, Session

# submodules
from scan_metadata_processor import ES_ENABLED, _
from scan_metadata_processor.utils.bouncer import exit_cli_error, exit_cli_success
from scan_metadata_processor.utils.proxies import proxy_settings

# #############################################################################
# ########## Globals ###############
# ##################################

# chronometer
START_TIME = default_timer()

# logs
logger = logging.getLogger(__name__)

# default CLI context.
# See: https://click.palletsprojects.com/en/7.x/commands/#context-defaults
CONTEXT_SETTINGS = dict(obj={})


# #############################################################################
# ####### Command-line ############
# #################################
@click.command()
@click.pass_context
def check(cli_context: click.Context):
    """Perform checks about requirements: folders, network..."""
    logger.info("CHECK started after {:5.2f}s.".format(default_timer() - START_TIME))

    # extract values from CLI context
    logs_folder = cli_context.obj.get("FOLDER_LOGS")
    output_folder = cli_context.obj.get("FOLDER_OUTPUT")

    # -- CHECK FOLDERS PERMISSIONS -----------------------------------------------------
    if not access(logs_folder, W_OK):
        err_msg = "The LOGS folder is not writable by '{}': {}".format(
            getlogin(), logs_folder.resolve()
        )
        exit_cli_error(err_msg)

    if not access(output_folder, W_OK):
        err_msg = "The OUTPUT folder is not writable by '{}': {}".format(
            getlogin(), output_folder.resolve()
        )
        exit_cli_error(err_msg)

    input_folder = Path(environ.get("ISOGEO_MD_PROCESSOR_INPUT_FOLDER"))
    if not access(input_folder, R_OK):
        err_msg = "The INPUT folder is not readable by '{}': {}".format(
            getlogin(), input_folder.resolve()
        )
        exit_cli_error(err_msg)

    # -- DATABASE ----------------------------------------------------------------------
    db_type = environ.get("ISOGEO_MD_PROCESSOR_DATABASE_TYPE", None)
    if db_type == "elastic-search" and not ES_ENABLED:
        err_msg = _(
            "Database type is set to Elastic Search but the required package is not installed."
        )
        exit_cli_error(err_msg)

    if not Path(environ.get("ISOGEO_MD_PROCESSOR_DATABASE_FOLDER")).is_dir():
        logger.error(
            "Folder path for the database is not correct: {}".format(
                environ.get("ISOGEO_MD_PROCESSOR_DATABASE_FOLDER")
            )
        )
    database_folder = Path(environ.get("ISOGEO_MD_PROCESSOR_DATABASE_FOLDER"))

    try:
        database_folder.mkdir(parents=True, exist_ok=True)
        logger.info("Database folder: {}".format(database_folder))
    except PermissionError as e:
        msg_err = (
            "Impossible to create the output folder. Does the user '{}' ({}) have write permissions "
            "on the OUTPUT folder?. Original error: {}".format(
                environ.get("userdomain"), getlogin(), e
            )
        )
        exit_cli_error(msg_err)

    if not IsogeoChecker().check_is_uuid(environ.get("ISOGEO_GROUP_ID")):
        exit_cli_error("Isogeo group UUID is not correct")

    # -- NETWORK -----------------------------------------------------------------------
    if int(environ.get("FULL_OFFLINE_MODE")) == 0 and db_type == "azure-isogeo":
        # load urls to check from text file defined in settings
        file_urls_to_check = Path(
            environ.get("URLS_TO_CHECK", "./urls_network_check.txt")
        )
        if file_urls_to_check.is_file():
            # urls_to_check = []
            with file_urls_to_check.open(mode="r", encoding="UTF8") as in_file:
                urls_to_check = in_file.read().splitlines()
            urls_to_check = [url for url in urls_to_check if len(url)]
        else:
            urls_to_check = None
        check_network(urls_to_check)
        logger.info("Network connections are good.")
    else:
        logger.info("Full offline mode enabled: network has not been tested.")

    # ending
    exit_cli_success(
        message=_("CHECK completed after {:5.2f}s.").format(
            default_timer() - START_TIME
        ),
        abort=False,
    )


def check_network(urls_to_check: list = None, check_azure_storage: bool = True) -> bool:
    """Check to contact URLs passed in as arg and a connection to Azure blob storage.

    :param list urls_to_check: list of URls to check. If None, fallback to Isogeo URLs. Defaults to: None - optional
    :param bool check_azure_storage: option to check also connection to Azure blob. Defaults to: True - optional

    :return: True
    :rtype: bool
    """
    fallback_urls = [
        "https://api.isogeo.com/about",
        "http://help.isogeo.com/",
        "https://id.api.isogeo.com/about",
    ]
    if urls_to_check is None:
        urls_to_check = fallback_urls

    # open http session
    with Session() as chuck:
        # set session options
        chuck.proxies = proxy_settings()
        chuck.verify = bool(int(environ.get("SSL_VERIFICATION", 1)))

        # try classic urls
        for url in urls_to_check:
            resp = chuck.get(url=url)
            try:
                resp.raise_for_status()
                logging.debug("URL check passed for: {}".format(url))
            except HTTPError as err:
                exit_cli_error(
                    _("Check URL failed. '{}' is not reachable. Trace: {}").format(
                        url, err
                    )
                )

    return True


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    # additionnal imports
    import multiprocessing

    # workaround for multiprocessing support for packaged version on Windows
    # see: https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Multiprocessing
    # and: https://stackoverflow.com/a/48805137/2556577
    multiprocessing.freeze_support()
    # launch cli
    check(obj={})



#! python3  # noqa E265

"""
    Manager for Elastic Search.

"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os import environ
from pathlib import Path

# 3rd party
from elasticsearch import Elasticsearch
from elasticsearch.exceptions import ConnectionError, RequestError

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# ##############################################################################
# ########## Classes ###############
# ##################################
class ElasticSearchManager(object):
    """Manage operations related to Elastic Search. Mainly a wrapper around ES Python client.\
        See: https://elasticsearch-py.readthedocs.io/en/master/

    :param str host: database host server. Defaults to: "localhost" - optional
    :param int port: database port. Defaults to: 9200 - optional
    :param str user: database user (with write rights). Defaults to: None - optional
    :param str password: user password. Defaults to: None - optional
    :param bool ssl: enable secured connection. Defaults to: True - optional
    :param int pool_size: number of connections to the database. Defaults to: 10 - optional
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 9200,
        user: str = None,
        password: str = None,
        ssl: bool = True,
        pool_size: int = 10,
    ):
        # store parameters as attributes
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.ssl = ssl
        self.pool_size = pool_size

    def connect(self) -> bool:
        """Establish connection to Elastic Search using module attributes."""
        # handle SSL options
        if self.ssl:
            self.scheme = "https"
        else:
            self.scheme = "http"

        # handle authentication
        if self.user is not None and self.password is not None:
            http_auth = (self.user, self.password)
        else:
            http_auth = None

        # try to connect
        self.es_client = None
        try:
            self.es_client = Elasticsearch(
                [{"host": self.host, "port": self.port,}],
                scheme=self.scheme,
                use_ssl=self.ssl,
                http_compress=True,
                http_auth=http_auth,
            )
        except ConnectionError as exc:
            logger.error(
                "Connection error. Are your settings ok? Traced error: {}".format(exc)
            )

        # check connection state
        if self.es_client.ping():
            logger.info(
                "Connection to Elastic Search succeeded: {}".format(
                    self.es_client.info()
                )
            )
            return True
        else:
            logger.error("Connection to Elastic Search failed.")
            return False

    def create_index(self, name: str = "isogeo", ignore_if_exists: bool = True):
        """Create the indice into the database using the defined mapping.

        :param str name: indice name. Defaults to: "isogeo" - optional
        :param bool ignore_if_exists: option to ignore existing index with same name. Defaults to: True - optional
        """
        try:
            if ignore_if_exists:
                self.es_client.indices.create(index=name, ignore=400)
            else:
                self.es_client.indices.create(index=name)
        except RequestError as exc:
            if ignore_if_exists and exc.error == "resource_already_exists_exception":
                logger.warning(
                    "Index '{}' already exists. Ignoring the error.".format(name)
                )
            else:
                logger.error(
                    "Index '{}' already exists. Traced error: {}".format(name, exc)
                )
                raise exc

        # store index name
        self.index_name = name

    def insert_document(
        self,
        document_to_index: dict,
        document_type: str = None,
        auto_refresh: str = "true",
    ):
        """Send documents to the Elastic indice.

        :param dict document_to_index: dictionary of document property to insert into database.
        :param str document_type: type of document to insert. Defaults to: None - optional
        :param bool auto_refresh: refresh just after the insertion. Valid choices: true, false, wait_for. Defaults to: "true" - optional
        """
        # insertion
        self.es_client.index(
            index=self.index_name,
            body=document_to_index,
            doc_type=document_type,
            refresh=auto_refresh,
        )


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    # additional imports
    from dotenv import load_dotenv

    # set log level
    logging.basicConfig(level=logging.INFO)
    # load dev environment variables
    if Path(".env").exists():
        load_dotenv(".env", override=True)

    elasticlient = ElasticSearchManager(
        host=environ.get("DATABASE_HOST"),
        port=environ.get("DATABASE_PORT"),
        pool_size=environ.get("DATABASE_POOL_CONNECTIONS"),
        user=environ.get("DATABASE_USER"),
        password=environ.get("DATABASE_PASSWORD"),
        ssl=False,
    )

    # connect
    elasticlient.connect()

    # ensure index is created
    elasticlient.create_index()

    # doc = {
    #     "author": "isogeo iioghjsfgkn",
    #     "abstract": "This is a sample testing metadata in a ES search engine",
    #     "title": "Sample metadata 2",
    #     "timestamp": datetime.now(),
    # }

    # # elasticlient.insert_document(document_to_index= )

    res = elasticlient.es_client.search(
        index=elasticlient.index_name, body={"query": {"match_all": {}}}
    )
    print("Got %d Hits:" % res["hits"]["total"]["value"])
    # for hit in res["hits"]["hits"]:
    #     print("%(timestamp)s %(author)s: %(title)s" % hit["_source"])



#! python3  # noqa: E265

"""
    Sub-command in charge of cleaning up temporary files.

    Author: Isogeo
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from datetime import datetime, timedelta
from os import environ
from pathlib import Path
from timeit import default_timer

# 3rd party library
import click
from send2trash import send2trash

# #############################################################################
# ########## Globals ###############
# ##################################

# chronometer
START_TIME = default_timer()

# logs
logger = logging.getLogger(__name__)

# default CLI context.
# See: https://click.palletsprojects.com/en/7.x/commands/#context-defaults
CONTEXT_SETTINGS = dict(obj={})


# #############################################################################
# ####### Command-line ############
# #################################
@click.command()
@click.option(
    "-rm",
    "--remove-empty-folders",
    help="Remove empty folders.",
    is_flag=True,
    default=True,
    show_default=True,
)
@click.pass_context
def clean(cli_context: click.Context, remove_empty_folders: bool):
    """Delete logs and output folders older than the frequency set in the configuration.

    \f
    :param click.core.Context cli_context: Click context
    :param bool remove_empty_folders: if passed, empty folders will be deleted too
    """
    logger.info(
        "Let's clean files and folders created more than {} days ago.".format(
            environ.get("ISOGEO_MD_PROCESSOR_CLEAN_FREQUENCY")
        )
    )
    date_ref_days_ago = datetime.now() - timedelta(
        days=int(environ.get("ISOGEO_MD_PROCESSOR_CLEAN_FREQUENCY"))
    )

    # logs folder
    li_folders_logs = Path(environ.get("ISOGEO_MD_PROCESSOR_LOGS")).iterdir()
    with click.progressbar(
        list(li_folders_logs), label="Removing log folders",
    ) as log_folders:
        for subobj in log_folders:
            # compare last modification date
            if datetime.fromtimestamp(subobj.stat().st_mtime) < date_ref_days_ago:
                logger.debug(
                    "CLEANER - LOGS - Detected folder/file outdated: {}".format(subobj)
                )
                try:
                    send2trash(str(subobj.resolve()))
                except Exception as e:
                    logger.warning(
                        "Unable to delete: {}. Original error: {}".format(
                            subobj.resolve(), e
                        )
                    )
            # handle empty folders
            if (
                remove_empty_folders
                and subobj.is_dir()
                and not list(subobj.glob("**/*"))
            ):
                logger.info(
                    "Empty folder spotted: {}. Sending it to the trash...".format(
                        str(subobj.resolve())
                    )
                )
                try:
                    send2trash(str(subobj.resolve()))
                except Exception as e:
                    logger.warning(
                        "Unable to delete: {}. Original error: {}".format(
                            subobj.resolve(), e
                        )
                    )

    # output folder
    li_folders_output = Path(environ.get("ISOGEO_MD_PROCESSOR_OUTPUT_FOLDER")).iterdir()
    with click.progressbar(
        list(li_folders_output), label="Removing output folders",
    ) as output_folders:
        for subobj in output_folders:
            if datetime.fromtimestamp(subobj.stat().st_ctime) < date_ref_days_ago:
                logger.info(
                    "CLEANER - OUTPUT- Detected folder/file outdated: {}".format(subobj)
                )
                try:
                    send2trash(str(subobj.resolve()))
                except Exception as e:
                    logger.warning(
                        "Unable to delete: {}. Original error: {}".format(
                            subobj.resolve(), e
                        )
                    )
            # handle empty folders
            if (
                remove_empty_folders
                and subobj.is_dir()
                and not list(subobj.glob("**/*"))
            ):
                logger.info(
                    "Empty folder spotted: {}. Sending it to the trash...".format(
                        str(subobj.resolve())
                    )
                )
                try:
                    send2trash(str(subobj.resolve()))
                except Exception as e:
                    logger.warning(
                        "Unable to delete: {}. Original error: {}".format(
                            subobj.resolve(), e
                        )
                    )


# #############################################################################
# ##### Stand alone program ########
# ##################################
if __name__ == "__main__":
    """Standalone execution."""
    # additionnal imports
    import multiprocessing

    # 3rd party
    from dotenv import load_dotenv

    # load settings as environment vars
    load_dotenv("./dev.env", override=True)
    # workaround for multiprocessing support for packaged version on Windows
    # see: https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Multiprocessing
    # and: https://stackoverflow.com/a/48805137/2556577
    multiprocessing.freeze_support()
    # launch cli
    clean(obj={})



#! python3  # noqa: E265

"""
    Command-line to read metadata produced by FME scripts of Isogeo Scan,
    convert it into Isogeo model, apply rules and export it to a database or Isogeo.

    Author: Isogeo
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from datetime import datetime
from logging.handlers import RotatingFileHandler
from os import environ, getlogin
from pathlib import Path
from platform import architecture
from platform import platform as opersys
from socket import gethostname
from timeit import default_timer
from urllib.request import getproxies

# 3rd party library
import click
from dotenv import load_dotenv

# submodules
from scan_metadata_processor import _
from scan_metadata_processor.__about__ import __version__
from scan_metadata_processor.commands import cli_check, cli_clean, cli_process
from scan_metadata_processor.utils.bouncer import exit_cli_error

# #############################################################################
# ########## Globals ###############
# ##################################

# chronometer
START_TIME = default_timer()

# logs
logger = logging.getLogger(__name__)

# default CLI context.
# See: https://click.palletsprojects.com/en/7.x/commands/#context-defaults
CONTEXT_SETTINGS = dict(obj={})

# #############################################################################
# ########## Functions #############
# ##################################


def log_start(log_filename="ISOGEO_MD_PROCESSOR_metadata_processor"):
    # required subfolder
    dir_logs = Path(environ.get("ISOGEO_MD_PROCESSOR_LOGS"))
    try:
        dir_logs.mkdir(exist_ok=True)
    except PermissionError as e:
        msg_err = (
            "Impossible to write the logs. Does the user '{}' ({}) have write permissions "
            "on the LOGS folder?. Original error: {}".format(
                environ.get("userdomain"), getlogin(), e
            )
        )
        exit_cli_error(msg_err)

    # create the logger
    logger = logging.getLogger()
    logging.captureWarnings(True)
    # logger.setLevel(logging.INFO)

    # set the format
    log_format = logging.Formatter(
        "%(asctime)s || %(levelname)s "
        "|| %(module)s - %(lineno)d ||"
        " %(funcName)s || %(message)s"
    )

    # log to the file
    log_file_handler = RotatingFileHandler(
        filename=dir_logs / "{}.log".format(log_filename),
        mode="a",
        maxBytes=3000000,
        backupCount=10,
        encoding="UTF-8",
    )

    log_file_handler.setLevel(logging.INFO)
    log_file_handler.setFormatter(log_format)

    # log to the console
    log_console_handler = logging.StreamHandler()
    log_console_handler.setLevel(logging.WARNING)
    log_console_handler.setFormatter(log_format)

    # add file and console handlers to the logger
    logger.addHandler(log_file_handler)
    # logger.addHandler(log_console_handler)

    # initialize the log
    logger.info(
        "===================== ISOGEO Scan Metadata Processor - Version {} =====================".format(
            __version__
        )
    )
    logger.info("Operating System: {}".format(opersys()))
    logger.info("Architecture: {}".format(architecture()[0]))
    logger.info("Computer: {}".format(gethostname()))
    logger.info("Launched by: {}".format(getlogin()))
    logger.info("OS Domain: {}".format(environ.get("userdomain")))
    logger.info("Network proxies detected: {}".format(len(getproxies())))

    return logger


# #############################################################################
# ####### Command-line ############
# #################################


@click.group(invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.option(
    "-l",
    "--label",
    help=_("Custom the name of the task (used to name the log file)"),
    default="isogeo_metadata_processor",
)
@click.option(
    "-s",
    "--settings",
    default=".env",
    help=_("Environment file containing settings"),
    type=click.Path(exists=True, readable=True),
)
@click.version_option(version=__version__, message="%(version)s")
@click.pass_context
def scan_metadata_processor(
    cli_context: click.Context, label: str, settings: Path,
):
    """Command-line checking settings and executing required operations.

    \f
    :param click.core.Context cli_context: Click context
    :param str label: name of run, used to custom some outputs (logs...)
    :param str settings: path to a settings file containing credentials to read database

    :Example:
        .. code-block:: powershell

        ./Isogeo_ScanMetadataProcessor.exe --label "Check" --settings ./settings.env check

    """
    # Load settings
    if Path(settings).exists():
        load_dotenv(settings, override=True)
        logging.info(
            "Settings loaded. Platfom used: {}".format(environ.get("ISOGEO_PLATFORM"))
        )
    else:
        click.Abort(
            FileNotFoundError("Settings file is missing. Please create an '.env' file.")
        )

    # create the log manager before all
    logger = log_start(label)

    # apply debug mode
    if int(environ.get("DEBUG")):
        logger.setLevel(logging.DEBUG)
        for handler in logger.handlers:
            handler.setLevel(logging.DEBUG)

    # create output folder
    output_folder = Path(
        environ.get("ISOGEO_MD_PROCESSOR_OUTPUT_FOLDER"),
        datetime.now().strftime("%Y%m%d_%H%M%S"),
    )
    try:
        output_folder.mkdir(parents=True, exist_ok=True)
    except PermissionError as e:
        msg_err = (
            "Impossible to create the output folder. Does the user '{}' ({}) have write permissions "
            "on the OUTPUT folder? Original error: {}".format(
                environ.get("userdomain"), getlogin(), e
            )
        )
        exit_cli_error(msg_err)

    # create logs subfolder
    logs_folder = Path(
        environ.get("ISOGEO_MD_PROCESSOR_LOGS"),
        datetime.now().strftime("%Y%m%d_%H%M%S"),
    )
    logs_folder.mkdir(parents=True, exist_ok=True)

    # save settings into the CLI context dict
    cli_context.obj["FOLDER_INPUT"] = Path(
        environ.get("ISOGEO_MD_PROCESSOR_INPUT_FOLDER")
    )
    cli_context.obj["FOLDER_OUTPUT"] = output_folder
    cli_context.obj["FOLDER_LOGS"] = logs_folder
    cli_context.obj["LABEL"] = label

    # db_manager.close_database()
    logger.info(
        _("Main CLI completed after {:5.2f}s.").format(default_timer() - START_TIME)
    )


# -- SUB-COMMANDS ----------------------------------------------------------------------
# Add subcommands to the main command group
scan_metadata_processor.add_command(cli_check.check)
scan_metadata_processor.add_command(cli_clean.clean)
scan_metadata_processor.add_command(cli_process.process)


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    # additionnal imports
    import multiprocessing

    # workaround for multiprocessing support for packaged version on Windows
    # see: https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Multiprocessing
    # and: https://stackoverflow.com/a/48805137/2556577
    multiprocessing.freeze_support()
    # launch cli
    scan_metadata_processor(obj={})



#! python3  # noqa E265

"""
    Isogeo Scan - Model of metadata output of Listing vector.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# standard library
import logging
import pprint
from typing import List


# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################
class Listing(object):
    """Output from FME scripts list-*."""

    ATTR_TYPES = {"listFeaturesTypes": list, "source": str}

    def __init__(self, results: List[dict], source: str = None, **kwargs):
        # default values for the object error/properties
        self._listFeatureTypes = None
        self._source = source

        # if values have been passed, so use them as objects error.
        # error are prefixed by an underscore '_'
        if results is not None:
            self._listFeatureTypes = [
                i.get("result").get("featureType") for i in results
            ]
        if source is not None:
            self._source = source

        # warn about unsupported attributes
        if len(kwargs):
            logger.warning(
                "Folllowings fields were not expected and have been ignored. "
                "Maybe consider adding them to the model: {}.".format(
                    " | ".join(kwargs.keys())
                )
            )

    # -- PROPERTIES --------------------------------------------------------------------
    # listFeaturesTypes
    @property
    def listFeaturesTypes(self) -> List[dict]:
        """Gets the list of feature types of the described dataset.

        :return: the list of feature types of the described dataset.
        :rtype: List[dict]
        """
        return self._listFeatureTypes

    @listFeaturesTypes.setter
    def listFeaturesTypes(self, listFeaturesTypes: List[dict]):
        """Sets the list of features types of the described dataset.

        :param List[dict] listFeaturesTypes: the list of feature types of the described dataset.
        """

        self._listFeatureTypes = listFeaturesTypes

    # source
    @property
    def source(self) -> str:
        """Gets the source.

        :return: the source of the listed feature types.
        :rtype: str
        """
        return self._source

    @source.setter
    def source(self, source: str):
        """Sets the source of the listed feature types.

        :param str source: the source of the listed feature types.
        """

        self._source = source

    # -- METHODS -----------------------------------------------------------------------
    def to_dict(self) -> dict:
        """Returns the model properties as a dict."""
        result = {}

        for attr, _ in self.ATTR_TYPES.items():
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value
        if issubclass(Listing, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model."""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other) -> bool:
        """Returns true if both objects are equal."""
        if not isinstance(other, Listing):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        """Returns true if both objects are not equal."""
        return not self == other


# ##############################################################################
# ##### Stand alone program ########
# ##################################
if __name__ == "__main__":
    """standalone execution."""



#! python3  # noqa: E265

"""
    Helper to handle Coordinate-System.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from typing import Tuple

# 3rd party
from isogeo_pysdk import CoordinateSystem
from scan_metadata_processor import Lookup

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)

# ##############################################################################
# ########## Classes ###############
# ##################################


def match_coordinate_system(
    group_srs: list, md_lookup: Lookup
) -> Tuple[int, CoordinateSystem]:
    """Determine if the dataset SRS is one of selected ones in the group configuration. \
    Match score:
      - 0: no SRS identified by lookup
      - 1: SRS identified and present among group's selection (4-4-2 as always)
      - 2: SRS identified but not selected in group configuration
      - 3: dataset seems to be projected but FME didn't not recognized the EPSG

    :param list group_srs: list of SRS selected in the group
    :param Lookup md_lookup: lookup model with a potential SRS EPSG Code

    :return: a tuple structured as (match_score, SRS or None)
    :rtype: Tuple[int, CoordinateSystem]
    """
    # variables
    out_coordsys = None

    # handle case where coordinate-system could not be identified during Scan
    if not isinstance(md_lookup.coordsys, dict):
        return (0, None)

    # extract information
    in_epsg_code = md_lookup.coordsys.get("EPSG")
    in_srs_name = md_lookup.coordsys.get("name")

    # match
    if in_epsg_code is not None and in_epsg_code.isdigit():
        # try match between group select SRS and lookup SRS
        srs_match = [d for d in group_srs if int(d.get("code")) == int(in_epsg_code)]
        if len(srs_match):
            logger.debug(
                "EPSG code matched group selected ESPG registry: {}.".format(
                    in_epsg_code
                )
            )
            # use Isogeo name instead
            out_coordsys = CoordinateSystem(
                code=in_epsg_code, name=srs_match[0].get("name")
            )
            has_srs = 1
        else:
            logger.debug(
                "EPSG code is not one of selected in group settings: {}.".format(
                    in_epsg_code
                )
            )
            # use SRS name found into lookup
            out_coordsys = CoordinateSystem(code=in_epsg_code, name=in_srs_name)
            has_srs = 2
    elif (
        "FME_0" in in_srs_name and in_srs_name == "" and md_lookup.envelope is not None
    ):
        logger.warning(
            "FME failed to recognize SRS but still calculated an envelope for: {}. "
            "Update your FME configuration.".format(md_lookup.name)
        )
        has_srs = 3
    else:
        # fallback
        has_srs = 0

    return (has_srs, out_coordsys)



#! python3  # noqa: E265

"""
    Metadata bout the package to easily retrieve informations about it.
    See: https://packaging.python.org/guides/single-sourcing-package-version/
"""

from datetime import date

__all__ = [
    "__author__",
    "__copyright__",
    "__email__",
    "__executable_name__",
    "__license__",
    "__package_name__",
    "__summary__",
    "__title__",
    "__uri__",
    "__version__",
]


__author__ = "Isogeo"
__copyright__ = "2019 - {0}, {1}".format(date.today().year, __author__)
__email__ = "contact@isogeo.com"
__executable_name__ = "Isogeo_ScanOffline.exe"
__license__ = "GNU Lesser General Public License v3.0"
__package_name__ = "scan_metadata_processor"
__pypi_name__ = "isogeo-scan-metadata-processor"
__summary__ = (
    "Command-line to read metadata produced by FME scripts of Isogeo Scan, "
    "convert it into Isogeo model, apply rules and export it to a database or Isogeo."
)
__title__ = "Isogeo Metadata Processor"
__title_clean__ = "".join(e for e in __title__ if e.isalnum())
__uri__ = "https://github.com/isogeo/scan-metadata-processor/"

__version__ = "1.1.0"
__version_info__ = tuple(
    [
        int(num) if num.isdigit() else num
        for num in __version__.replace("-", ".", 1).split(".")
    ]
)



#! python3  # noqa: E265

"""
    Read JSON files produced by Isogeo Scan.

    See:

    - https://docs.python.org/fr/3/library/JSON file.html

"""


# #############################################################################
# ########## Libraries #############
# ##################################
# standard library
import logging
import json
from json.decoder import JSONDecodeError
from io import BufferedIOBase
from os import access, R_OK
from pathlib import Path
from typing import Union

# 3rd party
from isogeo_pysdk import Metadata

# submodule
from scan_metadata_processor.parser import Listing, Lookup, Sign

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# ##############################################################################
# ########## Classes ###############
# ##################################
class MetadataJsonReader:
    """Read a JSON file produced by FME scripts of Isogeo Scan and return a Metadata.

    :param Union[str, Path, BufferedIOBase, bytes] in_json: path to the json file to read.
    """

    # global options to pass to open method
    # see: https://docs.python.org/fr/3/library/functions.html#open
    ENCODING_DEFAULT = "UTF-8"  # used as default encoding
    ENCODING_FALLBACK = "latin1"  # used as fallback encoding
    ENCODING_ERRORS_MODE = "replace"  # used to handle encoding errors

    def __init__(self, in_json: Union[str, Path, BufferedIOBase, bytes]):
        """Instanciating Isogeo Metadata JSON Reader."""
        # check and get JSON path
        if isinstance(in_json, (str, Path)):
            self.input_json = self.check_json_file(in_json)
            # extract data from input file
            with self.input_json.open(
                mode="r",
                encoding=self.ENCODING_DEFAULT,
                errors=self.ENCODING_ERRORS_MODE,
            ) as bytes_data:
                self.json_data = json.load(bytes_data)
        elif isinstance(in_json, BufferedIOBase):
            self.input_json = self.check_json_buffer(in_json)
            # extract data from input file
            self.json_data = json.load(bytes_data)
        elif isinstance(in_json, bytes):
            # extract data from input bytes
            self.json_data = json.loads(in_json)
        else:
            raise TypeError(
                "JSON type must be one of: str, Path, BufferedIOBase, bytes. Not: {}".format(
                    type(in_json)
                )
            )

        # guess source stage
        self.source

    # CHECKS
    def check_json_file(self, json_path: Union[str, Path]) -> Path:
        """Perform some checks on passed json file and load it as Path object.

        :param json_path: path to the json file to check

        :returns: sanitized json path
        :rtype: Path
        """
        # if path as string load it in Path object
        if isinstance(json_path, str):
            try:
                json_path = Path(json_path)
            except Exception as exc:
                raise TypeError("Converting json path failed: {}".format(exc))

        # check if file exists
        if not json_path.exists():
            raise FileExistsError(
                "JSON file to check doesn't exist: {}".format(json_path.resolve())
            )

        # check if it's a file
        if not json_path.is_file():
            raise IOError("JSON file is not a file: {}".format(json_path.resolve()))

        # check if file is readable
        if not access(json_path, R_OK):
            raise IOError("json file isn't readable: {}".format(json_path))

        # check integrity
        with json_path.open(
            mode="r", encoding=self.ENCODING_DEFAULT, errors=self.ENCODING_ERRORS_MODE
        ) as in_json_file:
            try:
                json.load(in_json_file)
            except JSONDecodeError as err:
                logger.error("JSON file is invalid: {}".format(json_path.resolve()))
                raise err
            except UnicodeDecodeError as err:
                logger.error(
                    "Failed to decode JSON file with encoding '{}': {}. "
                    "Trying to use the fallback encoding...".format(
                        self.ENCODING_DEFAULT, json_path.resolve()
                    )
                )
                # switch to fallback encoding
                self.ENCODING_DEFAULT = self.ENCODING_FALLBACK

                if self.ENCODING_DEFAULT != self.ENCODING_FALLBACK:
                    return self.check_json_file(json_path=json_path)

                raise err

        # return sanitized path
        return json_path

    def check_json_buffer(self, json_buffer: BufferedIOBase):
        """Perform some checks on passed json file and load it as Path object.

        :param json_buffer: path to the json file to check

        :returns: sanitized json path
        :rtype: Path
        """
        # check integrity
        try:
            json.load(json_buffer)
        except json.JSONDecodeError as exc:
            logger.error("Invalid JSON {}. Trace: {}".format(json_buffer, exc))
            raise exc

        # return sanitized path
        return json_buffer

    # PROPERTIES
    @property
    def source(self) -> str:
        """Look at the structure of the file and determine \
            at which stage of the scan it was produced or if it's an Isogeo export.

        :returns: scan stage ('listing', 'sign', 'lookup') or 'isogeo'
        :rtype: str
        """
        # read structure to identify source
        if isinstance(self.json_data, list) and "dataset" in self.json_data[0]:
            source = "lookup"
        elif isinstance(self.json_data, list) and "feature" in self.json_data[0]:
            source = "sign"
        elif isinstance(self.json_data, list) and "result" in self.json_data[0]:
            source = "listing"
        elif isinstance(self.json_data, dict):
            # test if it's already an Isogeo metadata
            try:
                Metadata.clean_attributes(self.json_data)
                source = "isogeo"
            except Exception:
                logger.warning(
                    "Unknown source: nor Scan, nor Isogeo. JSON data type (list expected): {}".format(
                        type(self.json_data)
                    )
                )
                source = None
        else:
            logger.warning(
                "Unknown stage. JSON data type (list expected): {}".format(
                    type(self.json_data)
                )
            )
            source = None

        # return sanitized path
        return source

    @property
    def as_model(self) -> object:
        """Load JSON into a model.

        :return: model corresponding to the parsed JSON content
        :rtype: object
        """
        if self.source == "listing":
            model = Listing(self.json_data)
        elif self.source == "lookup":
            model = Lookup(**self.json_data[0].get("dataset"))
        elif self.source == "sign":
            model = Sign(**self.json_data[0].get("feature"))
        elif self.source == "isogeo":
            model = Metadata.clean_attributes(self.json_data)
        else:
            model = None

        # return model
        return model


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    from os import environ
    from dotenv import load_dotenv

    logging.basicConfig(level=logging.DEBUG)
    # logger.setLevel(logging.INFO)

    # load dev environment vars
    load_dotenv("dev.env", override=True)
    fixtures_dir = Path(environ.get("FIXTURES_DIR"))

    for i in fixtures_dir.glob("**/*.json"):
        json_reader = MetadataJsonReader(i)

        # open JSON
        # print(json_reader.input_json, json_reader.source)
        # print(json_reader.input_json.name)
        model = json_reader.as_model
        if isinstance(model, Lookup):
            # print(model.isMapDocument)
            # print(model.isMapDocument)
            # print(model.isRaster)
            pass
        elif isinstance(model, Sign):
            # print(model.error, model.signatures)
            pass
        elif isinstance(model, Metadata):
            print(model.title_or_name())
        else:
            continue



#! python3  # noqa: E265

"""
    Manipulate zipped files.

    See:

    - https://docs.python.org/fr/3/library/zipfile.html

"""

# #############################################################################
# ########## Libraries #############
# ##################################

# standard library
import logging
from io import BufferedIOBase, BytesIO
from os import R_OK, access
from pathlib import Path
from typing import List, Union
from zipfile import BadZipFile, ZipFile, is_zipfile

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# ##############################################################################
# ########## Classes ###############
# ##################################
class Unzipper:
    """Check and extract files from a zip archive.

    :param input_zip: path to the zip file to unzip.
    """

    def __init__(self, input_zip: Union[str, Path]):
        """Instanciating Unzipper."""
        if isinstance(input_zip, (str, Path)):
            self.source_type = "file"
            self.zippo = self.check_zip(input_zip)
        elif isinstance(input_zip, BufferedIOBase):
            self.source_type = "buffer"
            self.zippo = self.check_zip(input_zip)
        else:
            raise TypeError(
                "Instance of input_zip must be one of: str, Path, BufferIOBase. Not: {}".format(
                    type(input_zip)
                )
            )

    @classmethod
    def check_zip(cls, input_zip: Union[str, Path]) -> Path:
        """Perform some checks on passed zip file and load it as Path object.

        :param input_zip: path to the zip file to unzip.

        :returns: sanitized zip path
        :rtype: Path
        """
        # if path as string load it in Path object
        if isinstance(input_zip, str):
            try:
                input_zip = Path(input_zip)
            except Exception as exc:
                raise TypeError("Converting zip path failed: {}".format(exc))

        if isinstance(input_zip, Path):
            # check if file exists
            if not input_zip.exists():
                raise FileExistsError(
                    "Zipfile to check doesn't exist: {}".format(input_zip.resolve())
                )

            # check if it's a file
            if not input_zip.is_file():
                raise IOError("Zipfile is not a file: {}".format(input_zip.resolve()))

            # check if file is readable
            if not access(input_zip, R_OK):
                raise IOError("Zip file isn't readable: {}".format(input_zip))

        elif isinstance(input_zip, BufferedIOBase):
            input_zip = BytesIO(input_zip.read())
        else:
            raise TypeError("Input zip type of instance is not correct.")

        # check if file is a zip
        if not is_zipfile(input_zip):
            raise BadZipFile("File isn't a ZIP: {}".format(input_zip))

        # check integrity
        try:
            with ZipFile(file=input_zip) as in_zip:
                in_zip.testzip()
        except BadZipFile as exc:
            raise exc

        # return sanitized path
        return input_zip

    def get_files_list(self) -> list:
        """Return the list of zipped files (just a wrapper around namelist() standard method).

        :return: list of zipped files
        :rtype: list
        """
        return ZipFile(self.zippo).namelist()

    def unzip(
        self,
        to_folder: Union[str, Path] = None,
        all_in: bool = True,
        only_extensions: tuple = None,
    ) -> list:
        """Extract files from zip. Can extract every files or filter on some extensions.

        :param Union[str, Path] to_folder: path to the output folder. \
        If it doesn't exist, it'll be created. Defaults to: None - optional
        :param bool all_in: option to extract all files. Defaults to: True - optional
        :param tuple only_extensions:  option to filter on certain extensions. \
        Defaults to: None - optional

        :return: list of extracted files
        :rtype: list

        :example:

        .. code-block:: python

            for i in fixtures_dir.glob("**/*.zip"):
                unzipper = Unzipper(i)
                print(unzipper.zipfile_path)

                # extract all files
                unzipper.unzip(fixtures_dir / "unzipper/all_in", all_in=1)

                # extract only JSON
                unzipper.unzip(
                    to_folder=Path(fixtures_dir / "unzipper/filtered"),
                    all_in=0,
                    only_extensions=(".json",),
                )

        """
        if self.source_type == "file":
            all_in = self._check_all_in(all_in, only_extensions)
            return self._unzip_to_disq(all_in, only_extensions, to_folder)
        elif self.source_type == "buffer":
            return self._unzip_to_memory(only_extensions=only_extensions)

    def _check_all_in(self, all_in: bool, only_extensions: tuple) -> bool:
        """Minimal checker of unzip parameters.

        :return: all_in value
        :rtype: bool
        """
        # check args conflicts
        if all_in and only_extensions is not None:
            logger.warning(
                "'all_in' and 'only_extension' are exclusive: "
                "if the first is True, then the second must be None. "
                "The option only_extension will be kept."
            )
            all_in = False
        elif not all_in and only_extensions is None:
            logger.warning(
                "'all_in' and 'only_extension' are exclusive: "
                "if the first is False, then the second must be a tuple. "
                "The option 'all_in' will be set to True."
            )
            all_in = True

        return all_in

    def _unzip_to_disq(
        self, all_in: bool, only_extensions: tuple, to_folder: Path
    ) -> list:
        """Sub-method of 'unzip', specilized for unzipping to disq.
        Same parameters of 'unzip'.
        """
        # ensure the the output folder is created
        try:
            to_folder.mkdir(exist_ok=True, parents=True)
        except PermissionError:
            raise PermissionError(
                "Output folder is not writable: {}".format(to_folder.resolve())
            )

        if not to_folder.is_dir():
            logger.warning(
                "Passed output path is not a folder: {}. Parent folder will be used instead.".format(
                    to_folder.resolve()
                )
            )
            to_folder = to_folder.parent

        # option to extract everything
        if all_in and only_extensions is None:
            with ZipFile(file=self.zippo) as in_zip:
                in_zip.extractall(to_folder)
            logger.info(
                "{} files extracted to {} from {}".format(
                    len(in_zip.namelist()), to_folder.resolve(), self.zippo.resolve(),
                )
            )
            # return with extracted files list
            return in_zip.namelist()

        # option to extract only certain files
        if not all_in and isinstance(only_extensions, tuple):
            li_extracted_files = []
            with ZipFile(file=self.zippo) as in_zip:
                for file_to_extract in in_zip.namelist():
                    if file_to_extract.endswith(only_extensions):
                        in_zip.extract(file_to_extract, to_folder)
                        li_extracted_files.append(file_to_extract)

            # return with extracted files list
            return li_extracted_files

    def _unzip_to_memory(self, only_extensions: tuple) -> List[bytes]:
        """Sub-method of unzip which unzip all files matching the extensions into a list of bytes.

        :return: list of JSONs as bytes
        :rtype: List[bytes]
        """
        li_out_files = []
        self.zippo.seek(0)
        with ZipFile(file=self.zippo) as in_zip:
            for file_in_zip in self.get_files_list():
                if file_in_zip.endswith(only_extensions):
                    li_out_files.append(in_zip.read(file_in_zip))

        return li_out_files


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    pass



#! python3  # noqa: E265

"""
    Compare metadata and calculate diffs.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from copy import copy
from typing import Tuple, Union

# 3rd party
from isogeo_pysdk import CoordinateSystem, FeatureAttribute, Metadata

# submodule
from scan_metadata_processor import _
from scan_metadata_processor.__about__ import __title__, __version__

# #############################################################################
# ########## Globals ###############
# ##################################

# log
logger = logging.getLogger(__name__)

# ##############################################################################
# ########## Classes ###############
# ##################################


class MetadataComparator:
    """Calculates diffs between two metadatas.

    :param Metadata metadata_old: reference metadata
    :param Metadata metadata_new: new metadata to compare with the reference
    """

    def __init__(self, metadata_old: Metadata, metadata_new: Metadata):
        """Instanciating module."""
        # store args as attributes
        self.md_old = copy(metadata_old)
        self.md_new = copy(metadata_new)

    def has_changed(self) -> bool:
        """Shortcut to determine if metadatas are different.

        :return: True if both metadata are different. False if not.
        :rtype: bool
        """
        return any(
            (
                self.diff_features_count(),
                self.diff_srs(),
                self.diff_features_attributes()[0],
            )
        )

    def diff_features_count(self) -> int:
        """Calculates the diff between features count. Only for vector datasets. \
        For non-vector data, returns 0. For vector datasets:
          - diff is positive if there is more objects than before.
          - diff is negative if there is less objects than before.

        :return: difference
        :rtype: int
        """
        if not all(i.type == "vectorDataset" for i in (self.md_old, self.md_new)):
            return 0

        # compare
        return self.md_new.features - self.md_old.features

    def diff_features_attributes(self) -> Tuple[int, set, set]:
        """Calculates the diff between features attributes. Only for vector datasets. \
        For non-vector data, returns (0,).

        :return: (diff, set of added attributes, set of removed attributes)
        :rtype: Tuple[0] or Tuple[int, set, set]
        """
        if not all(i.type == "vectorDataset" for i in (self.md_old, self.md_new)):
            return (0,)

        # shortcuts
        md_old_attrs = self.md_old.featureAttributes or []
        md_new_attrs = self.md_new.featureAttributes or []

        # ensure types
        self.md_old.featureAttributes = md_old_attrs = [
            FeatureAttribute(**f) if isinstance(f, dict) else f for f in md_old_attrs
        ]
        self.md_new.featureAttributes = md_new_attrs = [
            FeatureAttribute(**f) if isinstance(f, dict) else f for f in md_new_attrs
        ]

        # make feature attributes list unique
        md_new_attrs_set = set(x.name for x in md_new_attrs)
        md_old_attrs_set = set(x.name for x in md_old_attrs)

        # calculate diff between sets
        diff_set_add = md_new_attrs_set - md_old_attrs_set  # added
        diff_set_rmv = md_old_attrs_set - md_new_attrs_set  # removed
        return (len(diff_set_add) + len(diff_set_rmv), diff_set_add, diff_set_rmv)

    def diff_srs(self) -> int:
        """Calculates the diff between features count:
          - `-1`: coordinate-system have been removed
          - `0`: no change
          - `1`: coordinate-system have been added
          - `2`: coordinate-system have changed

        :return: difference code
        :rtype: int
        """
        # shortcuts
        md_old_srs = self.md_old.coordinateSystem
        md_new_srs = self.md_new.coordinateSystem

        # ensure input types
        if isinstance(md_old_srs, dict):
            md_old_srs = self.md_old.coordinateSystem = CoordinateSystem(**md_old_srs)
        if isinstance(md_new_srs, dict):
            md_new_srs = self.md_new.coordinateSystem = CoordinateSystem(**md_new_srs)

        # if both metadata have a coordinate-system, null or changed
        if isinstance(md_old_srs, CoordinateSystem) and isinstance(
            md_new_srs, CoordinateSystem
        ):
            if int(md_old_srs.code) == int(md_new_srs.code):
                return 0
            else:
                return 2
        # if previous version had a coordsys but not the new = removed
        elif isinstance(md_old_srs, CoordinateSystem) and not isinstance(
            md_new_srs, CoordinateSystem
        ):
            return -1
        # if previous version did not have a coordsys but the new = added
        elif not isinstance(md_old_srs, CoordinateSystem) and isinstance(
            md_new_srs, CoordinateSystem
        ):
            return 1
        else:
            return 0

    def diff_message(self) -> Union[str, None]:
        """Summarize diffences and returns a pre-formatted message. \
        Useful as update description. If no diff, returns None.

        :return: summary of diffs
        :rtype: str or None
        """
        # build output message
        out_msg = ""

        # -- FEATURES COUNT
        diff_features_count = self.diff_features_count()
        if diff_features_count > 0:
            out_msg += _("Entities count changed: ")
            out_msg += _("{} more objects (previously {})").format(
                abs(diff_features_count), self.md_old.features
            )
            out_msg += "\n"
        elif diff_features_count < 0:
            out_msg += _("Entities count changed: ")
            out_msg += _("{} objects less (previously {})").format(
                abs(diff_features_count), self.md_old.features
            )

            out_msg += "\n"
        else:
            pass

        # -- SRS
        diff_srs = self.diff_srs()
        if diff_srs < 0:
            out_msg += _(
                "\n**Coordinate-system** has been removed: {} ({}).".format(
                    self.md_old.coordinateSystem.name, self.md_old.coordinateSystem.code
                )
            )
            out_msg += "\n"
        elif diff_srs == 1:
            out_msg += _(
                "**Coordinate-system** has been added: {} ({}).".format(
                    self.md_new.coordinateSystem.name, self.md_new.coordinateSystem.code
                )
            )
            out_msg += "\n"
        elif diff_srs == 2:
            out_msg += _(
                "**Coordinate-system** have changed from {} ({}) to {} ({}).".format(
                    self.md_old.coordinateSystem.name,
                    self.md_old.coordinateSystem.code,
                    self.md_new.coordinateSystem.name,
                    self.md_new.coordinateSystem.code,
                )
            )
            out_msg += "\n"
        else:
            pass

        # -- FEATURE ATTRIBUTES
        diff_attrs = self.diff_features_attributes()
        if diff_attrs[0]:
            out_msg += _("\n**Feature attributes** have changed:\n")

            # added attrs
            if diff_attrs[1]:
                out_msg += _("\n- {} added: {}").format(
                    len(diff_attrs[1]), _("; ").join(diff_attrs[1])
                )

            else:
                out_msg += _("\n- no new attribute.")

            # removed attrs
            if diff_attrs[2]:
                out_msg += _("\n- {} removed: {}").format(
                    len(diff_attrs[2]), _("; ").join(diff_attrs[2])
                )

            else:
                out_msg += _("\n- no removed attribute.")

            out_msg += "\n"

        # if output message is still empty, returns None
        if out_msg == "":
            return None

        # add watermark
        out_msg += _("\n\n----\n*Processed by {} v{}*").format(__title__, __version__)

        return out_msg.strip()


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    pass



#! python3  # noqa: E265

# submodules
from .matchers import (  # noqa: F401
    MATCHER_FORMAT,  # noqa: F401
    MATCHER_GEOMETRY,  # noqa: F401
    match_coordinate_system,  # noqa: F401
)
from .lookup_to_isogeo import LookupToIsogeo  # noqa: F401



#! python3  # noqa E265

"""
    Isogeo Scan - Model of metadata output of lookup vector.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# standard library
import logging
import pprint

# #############################################################################
# ########## Globals ###############
# ##################################

logger = logging.getLogger(__name__)


# #############################################################################
# ########## Classes ###############
# ##################################
class Lookup(object):
    """Metadata produced by FME script lookup-*."""

    ATTR_TYPES = {
        "attributes": list,
        "bands": list,
        "bands_count": int,
        "bbox": list,
        "cols_count": int,
        "coordsys": dict,
        "dataSources": list,
        "description": str,
        "envelope": str,
        "fmeEnv": dict,
        "formatLong": str,
        "formatShort": str,
        "formatVersion": str,
        "geometryType": str,
        "name": str,
        "numberOfFeatures": int,
        "path": str,
        "rows_count": int,
    }

    ATTR_MAP = {
        "fme_formatShort": "formatShort",
        "fme_formatLong": "formatLong",
    }

    def __init__(
        self,
        attributes: list = None,
        bands: list = None,
        bands_count: int = None,
        bbox: list = None,
        cols_count: int = None,
        coordsys: dict = None,
        dataset: dict = None,
        dataSources: list = None,
        description: str = None,
        envelope: str = None,
        fmeEnv: dict = None,
        fme_formatShort: str = None,
        fme_formatLong: str = None,
        formatLong: str = None,
        formatShort: str = None,
        formatVersion: str = None,
        name: str = None,
        numberOfFeatures: int = None,
        path: str = None,
        rows_count: int = None,
        type: str = None,  # will be converted to geometry_type
        **kwargs,
    ):
        # default values for the object attributes/properties
        self._attributes = None
        self._bands = None
        self._bands_count = None
        self._bbox = None
        self._cols_count = None
        self._coordsys = None
        self._dataSources = None
        self._description = None
        self._envelope = None
        self._fmeEnv = None
        self._fme_formatShort = None
        self._fme_formatLong = None
        self._formatLong = None
        self._formatShort = None
        self._formatVersion = None
        self._geometryType = None
        self._name = None
        self._numberOfFeatures = None
        self._path = None
        self._rows_count = None

        # case when the JSON is passed
        if dataset:
            attributes = dataset.get("attributes")
            bands = dataset.get("bands")
            bands_count = dataset.get("bands_count")
            bbox = dataset.get("bbox")
            cols_count = dataset.get("cols_count")
            dataSources = dataset.get("dataSources")
            description = dataset.get("description")
            envelope = dataset.get("envelope")
            fmeEnv = dataset.get("fmeEnv")
            fme_formatShort = dataset.get("fme_formatShort")
            fme_formatLong = dataset.get("fme_formatLong")
            formatLong = dataset.get("formatLong")
            formatShort = dataset.get("formatShort")
            formatVersion = dataset.get("formatVersion")
            name = dataset.get("name")
            numberOfFeatures = dataset.get("numberOfFeatures")
            path = dataset.get("path")
            rows_count = dataset.get("rows_count")
            type = dataset.get("type")

        # if values have been passed, so use them as objects attributes.
        # attributes are prefixed by an underscore '_'
        if attributes is not None:
            self._attributes = attributes
        if bands is not None:
            self._bands = bands
        if bands_count is not None:
            self._bands_count = bands_count
        if bbox is not None:
            self._bbox = bbox
        if cols_count is not None:
            self._cols_count = cols_count
        if coordsys is not None:
            self._coordsys = coordsys
        if dataSources is not None:
            self._dataSources = dataSources
        if description is not None:
            self._description = description
        if envelope is not None:
            self._envelope = envelope
        if fmeEnv is not None:
            self._fmeEnv = fmeEnv
        # handle different formats names
        if fme_formatLong is not None:
            self.formatLong = fme_formatLong
        if fme_formatShort is not None:
            self.formatShort = fme_formatShort
        if formatLong is not None:
            self._formatLong = formatLong
        if formatShort is not None:
            self._formatShort = formatShort
        if formatVersion is not None:
            self._formatVersion = formatVersion
        if name is not None:
            self._name = name
        if numberOfFeatures is not None:
            self._numberOfFeatures = numberOfFeatures
        if path is not None:
            self._path = path
        if rows_count is not None:
            self._rows_count = rows_count
        if type is not None:
            self._geometryType = type

        # warn about unsupported attributes
        if len(kwargs):
            logger.warning(
                "Folllowing fields were not expected and have been ignored. "
                "Maybe consider adding them to the model: {}. Object name: {}".format(
                    " | ".join(kwargs.keys()), name
                )
            )

    # -- PROPERTIES --------------------------------------------------------------------
    # attributes
    @property
    def attributes(self) -> int:
        """Gets the attributes.

        :return: The attributes.
        :rtype: int
        """
        return self._attributes

    @attributes.setter
    def attributes(self, attributes: int):
        """Sets the attributes.

        :param int attributes: The attributes.
        """

        self._attributes = attributes

    # bands properties
    @property
    def bands(self) -> list:
        """Gets the bands.

        :return: The bands.
        :rtype: list
        """
        return self._bands

    @bands.setter
    def bands(self, bands: list):
        """Sets the bands.

        :param list bands: The bands.
        """

        self._bands = bands

    # bands count
    @property
    def bands_count(self) -> list:
        """Gets the bands_count.

        :return: The bands_count.
        :rtype: list
        """
        return self._bands_count

    @bands_count.setter
    def bands_count(self, bands_count: list):
        """Sets the bands_count.

        :param list bands_count: The bands_count.
        """

        self._bands_count = bands_count

    # Bounding Box
    @property
    def bbox(self) -> dict:
        """Gets the bbox of this Metadata.

        :return: The bbox of this Metadata.
        :rtype: dict
        """
        return self._bbox

    @bbox.setter
    def bbox(self, bbox: dict):
        """Sets the coordinate systems of this Metadata.

        :param dict bbox: to be set
        """

        self._bbox = bbox

    # cols count
    @property
    def cols_count(self) -> list:
        """Gets the cols_count.

        :return: The cols_count.
        :rtype: list
        """
        return self._cols_count

    @cols_count.setter
    def cols_count(self, cols_count: list):
        """Sets the cols_count.

        :param list cols_count: The cols_count.
        """

        self._cols_count = cols_count

    # coordinateSystem
    @property
    def coordsys(self) -> dict:
        """Gets the coordsys of this Metadata.

        :return: The coordsys of this Metadata.
        :rtype: dict
        """
        return self._coordsys

    @coordsys.setter
    def coordsys(self, coordsys: dict):
        """Sets the coordinate systems of this Metadata.

        :param dict coordsys: to be set
        """

        self._coordsys = coordsys

    # related data sources
    @property
    def dataSources(self) -> list:
        """Gets the dataSources.

        :return: The dataSources of this Metadata.
        :rtype: list
        """
        return self._dataSources

    @dataSources.setter
    def dataSources(self, dataSources: list):
        """Sets the dataSources.

        :param list dataSources: the dataSources of this Metadata.
        """

        self._dataSources = dataSources

    # metadata description
    @property
    def description(self) -> str:
        """Gets the description.

        :return: The description of this Metadata.
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description: str):
        """Sets the description used into Isogeo filters of this Metadata.

        :param str description: the description of this Metadata.
        """

        self._description = description

    # search envelope
    @property
    def envelope(self) -> dict:
        """Gets the abilities.

        :return: The abilities.
        :rtype: dict
        """
        return self._envelope

    @envelope.setter
    def envelope(self, envelope: dict):
        """Sets the envelope of this InlineResponse2001.

        The aggregated convex hull for the entire serach numberOfFeatures. \
        Might be null (if the numberOfFeatures span the entire globe, for instance).  # noqa: E501

        :param envelope: The envelope of this InlineResponse2001.  # noqa: E501
        :type: object
        """

        self._envelope = envelope

    # FME Environment
    @property
    def fmeEnv(self) -> dict:
        """Gets the modified.

        :return: The modified.
        :rtype: dict
        """
        return self._fmeEnv

    @fmeEnv.setter
    def fmeEnv(self, fmeEnv: dict):
        """Sets the fmeEnv of this InlineResponse2001.

        :param fmeEnv: The fmeEnv of this InlineResponse2001.  # noqa: E501
        :type: object
        """

        self._fmeEnv = fmeEnv

    # formatShort UUID
    @property
    def formatShort(self) -> str:
        """Gets the formatShort.

        :return: The formatShort.
        :rtype: str
        """
        return self._formatShort

    @formatShort.setter
    def formatShort(self, formatShort: str):
        """Sets the formatShort.

        :param str formatShort: The formatShort.
        """

        self._formatShort = formatShort

    # search last modification date
    @property
    def formatLong(self) -> str:
        """Gets the modified.

        :return: The modified.
        :rtype: str
        """
        return self._formatLong

    @formatLong.setter
    def formatLong(self, formatLong: str):
        """Sets the formatLong.

        :param str formatLong: The formatLong
        """

        self._formatLong = formatLong

    @property
    def formatVersion(self) -> str:
        """Gets the format version.

        :return: the format version.
        :rtype: str
        """
        return self._formatVersion

    @formatVersion.setter
    def formatVersion(self, formatVersion: str):
        """Sets the formatVersion.

        :param str formatVersion: format version
        """

        self._formatVersion = formatVersion

    # geometry type = type
    @property
    def geometryType(self) -> str:
        """Gets the type of geometry of the described dataset. \
        It's mapped from the `type` attribute in Scan output.

        :return: the type of geometry of the described dataset.
        :rtype: str
        """
        return self._geometryType

    @geometryType.setter
    def geometryType(self, geometryType: str):
        """Sets the type of geometry of the described dataset.

        :param str geometryType: the type of geometry of the described dataset.
        """

        # check type value
        self._geometryType = geometryType

    # search creation date
    @property
    def name(self) -> int:
        """Gets the created.

        :return: The created.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: int):
        """Sets the name of this InlineResponse2001.

        :param int name: The name
        """

        self._name = name

    # search numberOfFeatures path
    @property
    def numberOfFeatures(self) -> int:
        """Gets the tag.

        :return: The tag.
        :rtype: int
        """
        return self._numberOfFeatures

    @numberOfFeatures.setter
    def numberOfFeatures(self, numberOfFeatures: int):
        """Sets the numberOfFeatures of this InlineResponse2001.

        :param int numberOfFeatures: The numberOfFeatures of this InlineResponse2001.  # noqa: E501
        """

        self._numberOfFeatures = numberOfFeatures

    # path
    @property
    def path(self) -> dict:
        """Gets the path.

        :return: The path.
        :rtype: dict
        """
        return self._path

    @path.setter
    def path(self, path: dict):
        """Sets the path of this InlineResponse2001.

        The aggregated set of path for the entire search numberOfFeatures  # noqa: E501

        :param path: The path of this InlineResponse2001.  # noqa: E501
        :type: path
        """

        self._path = path

    # rows count
    @property
    def rows_count(self) -> list:
        """Gets the rows_count.

        :return: The rows_count.
        :rtype: list
        """
        return self._rows_count

    @rows_count.setter
    def rows_count(self, rows_count: list):
        """Sets the rows_count.

        :param list rows_count: The rows_count.
        """

        self._rows_count = rows_count

    # -- SPECIFIC TO IMPLEMENTATION ----------------------------------------------------
    @property
    def isMapDocument(self) -> bool:
        """Determine if the metadata is about a map document containing datasets\
             (like *.mxd, *.lyr, etc.) and not a single dataset."""
        if self._dataSources is not None:
            return True
        else:
            return False

    @property
    def isRaster(self) -> bool:
        """Determine if the metadata describes a raster dataset."""
        if (
            self._bands is not None
            and not self._numberOfFeatures
            and not self.isMapDocument
        ):
            return True
        else:
            return False

    @property
    def isVector(self) -> bool:
        """Determine if the metadata describes a vector dataset."""
        if not self._bands_count and self._numberOfFeatures and not self.isMapDocument:
            return True
        else:
            return False

    # -- METHODS -----------------------------------------------------------------------
    def to_dict(self) -> dict:
        """Returns the model properties as a dict."""
        result = {}

        for attr, _ in self.ATTR_TYPES.items():
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(
                    map(lambda x: x.to_dict() if hasattr(x, "to_dict") else x, value)
                )
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(
                    map(
                        lambda item: (item[0], item[1].to_dict())
                        if hasattr(item[1], "to_dict")
                        else item,
                        value.items(),
                    )
                )
            else:
                result[attr] = value
        if issubclass(Lookup, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model."""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other) -> bool:
        """Returns true if both objects are equal."""
        if not isinstance(other, Lookup):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other) -> bool:
        """Returns true if both objects are not equal."""
        return not self == other


# ##############################################################################
# ##### Stand alone program ########
# ##################################
if __name__ == "__main__":
    """standalone execution."""
    sample = Lookup()



#! python3  # noqa: E265

import warnings

# submodules
from .__about__ import __version__  # noqa: F401
from .utils.localizer import Localizer  # must be imported before others

# instanciate the translator
localizer = Localizer()
_ = localizer.tr

# subpackages
from .comparator import MetadataComparator  # noqa: E402 F401
from .parser import Listing, Lookup, Sign  # noqa: E402 F401
from .converter import (  # noqa: E402 F401
    MATCHER_FORMAT,
    MATCHER_GEOMETRY,
    LookupToIsogeo,
    match_coordinate_system,
)
from .parser import MetadataJsonReader, Unzipper  # noqa: E402 F401

# must be imported at last
from .database.isogeo import MetadataSynchronizer  # noqa: E402 F401

# conditional import
try:
    from elasticsearch import Elasticsearch  # noqa: F401

    ES_ENABLED = True
    from .database import ElasticSearchManager  # noqa: F401
except ImportError:
    ES_ENABLED = False
    warnings.warn(
        message="ElasticSearch package is not installed. Related functions cannot be used.",
        category=ImportWarning,
    )



#! python3  # noqa: E265

# submodules
from .matcher_geometry_types import MATCHER_GEOMETRY  # noqa: F401
from .matcher_formats import MATCHER_FORMAT  # noqa: F401
from .matcher_srs import match_coordinate_system  # noqa: F401



# coding: utf-8
#! python3  # noqa: E265

"""
    Minimalist slugifier.

    Author: Isogeo
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
import re
import unicodedata

# #############################################################################
# ########## Globals ###############
# ##################################

# logs
logger = logging.getLogger(__name__)

# for slugified
_regex_slugify_strip = re.compile(r"[^\w\s-]")
_regex_slugify_hyphenate = re.compile(r"[-\s]+")

# #############################################################################
# ########## Functions #############
# ##################################


def sluggy(text_to_slugify: str, replacer: str = "-") -> str:
    """Very basic slugifier using only Python Standard Library.

    :param str text_to_slugify: text to slugify

    :return: slug
    :rtype: str

    :example:

    .. code-block:: python

        sample_txt = "Oyé oyé brâves gens de 1973 ! Hé oh ! Sentons-nous l'ail %$*§ ?!"
        print(sluggy(sample_txt))
        > oye-oye-braves-gens-de-1973-he-oh-sentons-nous-lail

    """
    slug = (
        unicodedata.normalize("NFKD", text_to_slugify)
        .encode("ascii", "ignore")
        .decode("ascii")
    )
    slug = _regex_slugify_strip.sub("", slug).strip().lower()
    slug = _regex_slugify_hyphenate.sub(replacer, slug)

    return slug


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    sample_txt = "Oyé oyé brâves gens de 1973 ! Hé oh ! Sentons-nous l'ail %$*§ ?!"
    print(sluggy(sample_txt))



#! python3  # noqa: E265

"""
    Model converter from lookup to Isogeo model.
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import json
import logging
import re
from time import gmtime, strftime
from typing import List

# 3rd party
import geojson
from isogeo_pysdk import CoordinateSystem, Event, FeatureAttribute, Metadata, Workgroup

# submodules
from scan_metadata_processor import _
from scan_metadata_processor.__about__ import __title__, __version__
from scan_metadata_processor.converter import (
    MATCHER_FORMAT,
    MATCHER_GEOMETRY,
    match_coordinate_system,
)
from scan_metadata_processor.parser.models import Lookup

# #############################################################################
# ########## Globals ###############
# ##################################

# log
logger = logging.getLogger(__name__)

# regex
reg_digits_parentesis = re.compile(r"^.*?\([^\d]*(\d+)[^\d]*\).*$")

# ##############################################################################
# ########## Classes ###############
# ##################################


class LookupToIsogeo:
    """Model converter from lookup to Isogeo model.

    :param Lookup in_lookup: source lookup model
    :param list isogeo_formats: Isogeo formats registry
    :param Workgroup isogeo_group: Isogeo group
    :param list group_coordsys: list of group selected coordinate-systems. Optional. Defaults to [].
    """

    COORDINATES_ROUND_LEVEL = 3

    def __init__(
        self,
        in_lookup: Lookup,
        isogeo_formats: list,
        isogeo_group: Workgroup,
        group_coordsys: List = [],
    ):
        """Instanciation method."""
        self.in_lookup = in_lookup
        self.formats = isogeo_formats
        self.group = isogeo_group
        self.group_srs = group_coordsys

    def prepare_event(
        self, evt_type: str = "creation", evt_description: str = None
    ) -> Event:
        """Prepare event text and date.

        :param str evt_type: type of event to create. Defaults to: "creation" - optional
        :param str evt_description: event text. Defaults to: None - optional

        :return: event object ready to be created
        :rtype: Event
        """
        # prepare text
        if evt_type == "creation":
            evt_kind = "update"
            # default text
            default_txt = _(
                "Metadata generated by the Scan, then created by {} v{}."
            ).format(__title__, __version__,)

            # custom or default text
            if evt_description:
                evt_txt = evt_description
            else:
                evt_txt = default_txt

            if self.in_lookup.fmeEnv:
                evt_txt += "\nUsed ETL: {} ({}.fmw)".format(
                    self.in_lookup.fmeEnv.get("version"),
                    self.in_lookup.fmeEnv.get("fmw"),
                )
        elif evt_type == "recreation":
            evt_kind = "update"
            # default text
            default_txt = _(
                "The metadata has been automatically recreated after it was "
                "manually deleted because the source dataset was scanned again."
            )
            # custom or default text
            if evt_description:
                evt_txt = evt_description
            else:
                evt_txt = default_txt
        elif evt_type == "update":
            evt_kind = "update"
            # default text
            default_txt = _(
                "Updated because source dataset has been detected as modified."
            )
            # custom or default text
            if evt_description:
                evt_txt = evt_description
            else:
                evt_txt = default_txt
        else:
            pass

        # creation event
        evt_date = strftime("%Y-%m-%d", gmtime())

        return Event(date=evt_date, description=evt_txt, kind=evt_kind)

    def prepare_feature_attributes(self) -> List[FeatureAttribute]:
        """Prepare feature attributes to be added to the metadata (only for vectors).

        :return: list of feature attributes ready to be added
        :rtype: List[FeatureAttribute]
        """
        out_feature_attributes = []

        # if no vector, no feature attributes
        if self.match_type != "vectorDataset":
            return out_feature_attributes

        # check if attributes were detected by the scan
        if not isinstance(self.in_lookup.attributes, list) or not len(
            self.in_lookup.attributes
        ):
            return out_feature_attributes

        for featattr in self.in_lookup.attributes:
            # extract length from feature-attribute type
            length_filter = reg_digits_parentesis.match(featattr.get("type"))
            if length_filter:
                attr_length = length_filter.group(1)
                attr_type = featattr.get("type").replace("({})".format(attr_length), "")
            else:
                attr_length = None
                attr_type = featattr.get("type")

            # load in object and to the output list
            out_feature_attributes.append(
                FeatureAttribute(
                    name=featattr.get("name"), dataType=attr_type, length=attr_length
                )
            )

        return out_feature_attributes

    def extract_envelope(self) -> dict:
        """Load envelope as GeoJSON, validate it and reduce precision to be more light.

        :return: geographic envelope as dict
        :rtype: dict
        """
        out_envelope = None

        if self.in_lookup.envelope:
            try:
                in_envelope = geojson.loads(self.in_lookup.envelope)
                if not in_envelope.is_valid:
                    logger.warning(
                        "Envelope GeoJSON of '{}' is not valid: {}".format(
                            self.in_lookup.name, in_envelope.errors()
                        )
                    )
                # round coordinates
                out_envelope = geojson.GeoJSON.to_instance(
                    geojson.utils.map_coords(
                        lambda x: round(x, self.COORDINATES_ROUND_LEVEL), in_envelope
                    )
                )
            except Exception as err:
                logger.error(
                    "Failed loading the envelope from {}. Using the raw string."
                    "Trace: {}".format(self.in_lookup.name, err)
                )
                out_envelope = json.loads(self.in_lookup.envelope)

        return out_envelope

    def extract_name(self) -> str:
        """Extract dataset name from lookup. \
        Handle specific cases like FileGDB Open which adds an ugly slash.

        :rtype: str
        """
        out_name = self.in_lookup.name

        # using open api driver add a forbidden char: '/'. Let's remove it.
        if (
            self.match_format_code == "filegdb"
            and self.in_lookup.formatShort == "FILEGDB"
        ):
            out_name = out_name.replace(r"/", ".")

        return out_name

    def extract_path(self) -> str:
        """Extract path.

        :return: dataset path
        :rtype: str
        """
        out_path = None
        if self.in_lookup.path:
            out_path = self.in_lookup.path

        return out_path

    @property
    def match_srs(self) -> CoordinateSystem:
        """If group coordinate-system has been passed to the class, use the matcher. \
        Otherwise, just return a CoordinateSystem with the lookup EPSG code \
        or None missing.

        :return: Isogeo coordinate system
        :rtype: CoordinateSystem
        """
        return match_coordinate_system(self.group_srs, self.in_lookup)[1]

    @property
    def match_format_code(self) -> str:
        """Try to match lookup format (FME short and long names) with:
          1. custom matching table included in submodule
          2. Isogeo formats registry codes.

        :return: Isogeo format code
        :rtype: str
        """
        isogeo_formats_codes = [i.get("code") for i in self.formats]
        # matching format
        md_format = MATCHER_FORMAT.get(self.in_lookup.formatShort)
        # first try to refer to Isogeo formats matrix
        if md_format in isogeo_formats_codes:
            logger.debug(
                "Format '{}' matched custom formats conversion table: {}".format(
                    self.in_lookup.formatShort, md_format
                )
            )
            out_format_code = md_format
        elif (
            isinstance(self.in_lookup.formatShort, str)
            and self.in_lookup.formatShort.lower() in isogeo_formats_codes
        ):
            logger.debug(
                "Format '{}' found in Isogeo formats registry (codes): {}".format(
                    self.in_lookup.formatShort.lower(), md_format
                )
            )
            out_format_code = self.in_lookup.formatShort.lower()
        else:
            out_format_code = None
            logging.warning(
                "Format '{}' not recognized for file: {}".format(
                    self.in_lookup.formatShort, self.in_lookup.name
                )
            )

        return out_format_code

    @property
    def match_format_version(self) -> str:
        """Try to determine the format version when there is only one.

        :return: format version
        :rtype: str
        """
        out_format_version = None

        # without format code, no format version
        if not self.match_format_code:
            return out_format_version

        # get the Isogeo format matching format code
        matched_fmt = [
            fmt for fmt in self.formats if fmt.get("code") == self.match_format_code
        ][0]

        # remove null versions
        matched_versions = [
            version for version in matched_fmt.get("versions") if version is not None
        ]

        # if only one version,so let's use it!
        if len(matched_versions) == 1:
            out_format_version = matched_versions[0]

        return out_format_version

    @property
    def match_geometry_type(self) -> str:
        """Match geometry type from lookup (FME) with Isogeo.

        :return: dataset geometry type
        :rtype: str
        """
        # matching format
        geometry_type = MATCHER_GEOMETRY.get(self.in_lookup.geometryType)

        return geometry_type

    @property
    def match_type(self) -> str:
        """Match dataset type, mainly using the lookup model included helpers.

        :return: metadata type
        :rtype: str
        """
        # default type
        out_md_type = "resource"

        # match between lookup status and type
        if self.in_lookup.isVector:
            out_md_type = "vectorDataset"
        elif self.in_lookup.isRaster:
            out_md_type = "rasterDataset"

        return out_md_type

    def as_metadata(self) -> Metadata:
        """Use submethods to return the lookup as an Isogeo metadata model \
        ready to be created.

        :return: Isogeo metadata converted from source lookup
        :rtype: Metadata
        """
        consolidated_md = Metadata(
            _creator=self.group,
            coordinateSystem=self.match_srs,
            editionProfile="manual",
            envelope=self.extract_envelope(),
            features=self.in_lookup.numberOfFeatures or None,
            featureAttributes=self.prepare_feature_attributes(),
            format=self.match_format_code,
            formatVersion=self.match_format_version,
            geometry=self.match_geometry_type,
            language=self.group.metadataLanguage or "fr",
            name=self.extract_name(),
            path=self.extract_path(),
            title=self.extract_name(),
            series=False,
            type=self.match_type,
        )

        return consolidated_md


# #############################################################################
# ##### Main #######################
# ##################################
if __name__ == "__main__":
    pass



# coding: utf-8
#! python3  # noqa: E265

"""
    Command-line to run Scan scripts without connection and post-synchronize the results.

    Author: Isogeo
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os import environ
from timeit import default_timer
from urllib.request import getproxies

# #############################################################################
# ########## Globals ###############
# ##################################

# chronometer
START_TIME = default_timer()

# logs
logger = logging.getLogger(__name__)

# default CLI context.
# See: https://click.palletsprojects.com/en/7.x/commands/#context-defaults
CONTEXT_SETTINGS = dict(obj={})

# #############################################################################
# ########## Functions #############
# ##################################


def proxy_settings():
    """Retrieves network proxy settings from OS or an environment file."""
    if environ.get("HTTP_PROXY") or environ.get("HTTPS_PROXY"):
        proxy_settings = {
            "http": environ.get("HTTP_PROXY"),
            "https": environ.get("HTTPS_PROXY"),
        }
        logger.debug(
            "Proxies settings found in environment vars (loaded from .env file): {}".format(
                proxy_settings
            )
        )
    elif getproxies():
        proxy_settings = getproxies()
        logger.debug("Proxies settings found in the OS: {}".format(proxy_settings))
    else:
        logger.debug("No proxy settings found in environment vars nor OS settings.")
        proxy_settings = None

    return proxy_settings


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    pass



#! python3  # noqa: E265


"""
    Setup script to package into a Python module
"""

# ############################################################################
# ########## Libraries #############
# ##################################

# standard library
import pathlib

# 3rd party
from setuptools import find_packages, setup

# package (to get version)
from scan_metadata_processor import __about__

# ############################################################################
# ########### Globals ##############
# ##################################

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# ############################################################################
# ########## Packaging #############
# ##################################
setup(
    # metadata
    name=__about__.__pypi_name__,
    version=__about__.__version__,
    author=__about__.__author__,
    author_email=__about__.__email__,
    description=__about__.__summary__,
    long_description=README,
    long_description_content_type="text/markdown",
    keywords="Isogeo scan JSON middleware",
    url=__about__.__uri__,
    project_urls={
        "Docs": "http://help.isogeo.com/scan/isogeo-scan-metadata-processor/",
        "Bug Reports": "{}issues/".format(__about__.__uri__),
        "Source": __about__.__uri__,
    },
    # implementation
    python_requires=">=3.6, <4",
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    # packaging
    py_modules=["scan_metadata_processor"],
    packages=find_packages(
        exclude=["contrib", "docs", "*.tests", "*.tests.*", "tests.*", "tests"]
    ),
    include_package_data=True,
    install_requires=[
        "click==7.1.*",
        "geojson>=2.5,<2.6",
        "isogeo-pysdk>=3.4,<3.5",
        "python-dotenv>=0.12,<0.14",
        "semver>=2.9,<2.11",
        "Send2Trash==1.5.*",
    ],
    extras_require={
        "dev": ["black", "flake8"],
        "db-elastic": ["elasticsearch>=7.5,<7.7"],
        "test": ["pytest", "pytest-cov"],
    },
    entry_points="""
        [console_scripts]
        scan-metadata-processor=scan_metadata_processor.cli:scan_metadata_processor
    """,
)



#! python3  # noqa: E265

"""
    Sub-command in charge of processing input ZIP files and loading results into database.

    Author: Isogeo
"""

# #############################################################################
# ########## Libraries #############
# ##################################

# Standard library
import logging
from os import environ, rename
from pathlib import Path
from timeit import default_timer

# 3rd party library
import click

# submodules
from scan_metadata_processor import ES_ENABLED, MetadataJsonReader, Unzipper, _
from scan_metadata_processor.utils.bouncer import (
    exit_cli_error,
    exit_cli_normal,
    exit_cli_success,
)

if ES_ENABLED:
    from scan_metadata_processor.database.elasticsearch import ElasticSearchManager

# #############################################################################
# ########## Globals ###############
# ##################################

# chronometer
START_TIME = default_timer()

# logs
logger = logging.getLogger(__name__)

# default CLI context.
# See: https://click.palletsprojects.com/en/7.x/commands/#context-defaults
CONTEXT_SETTINGS = dict(obj={})


# #############################################################################
# ####### Command-line ############
# #################################
@click.command()
@click.pass_context
def process(cli_context: click.Context):
    """Process input metadata (zipped) and load into database."""

    logger.info("PROCESS started after {:5.2f}s.".format(default_timer() - START_TIME))

    # retrieve values from CLI context
    input_folder = cli_context.obj.get("FOLDER_INPUT")
    output_folder = cli_context.obj.get("FOLDER_OUTPUT")

    # -- ZIP EXTRACTION ----------------------------------------------------------------
    # looking for zip files
    extraction_folder = Path(output_folder / "to_process")
    li_zip_files = list(input_folder.glob("**/*.zip"))
    if not len(li_zip_files):
        exit_cli_normal("No ZIP files found in: {}".format(extraction_folder.resolve()))

    # parse zip files and extract JSON files
    with click.progressbar(
        iterable=li_zip_files, length=len(li_zip_files), label="Extracting ZIP files..."
    ) as prgbar_zip_to_extract:
        for in_zip in prgbar_zip_to_extract:
            unzipper = Unzipper(in_zip)
            extracted = unzipper.unzip(
                to_folder=extraction_folder, all_in=0, only_extensions=(".json",),
            )
            if not extracted:
                logger.warning("No JSON file found into the ZIP: {}".format(in_zip))
                continue
            else:
                logger.debug(
                    "{} JSON files extracted from '{}' to '{}'".format(
                        len(extracted), in_zip, extraction_folder
                    )
                )

    # -- DATABASE CONFIGURATION --------------------------------------------------------
    db_type = environ.get("ISOGEO_MD_PROCESSOR_DATABASE_TYPE", None)
    if db_type is None:
        exit_cli_error(
            "setting ISOGEO_MD_PROCESSOR_DATABASE_TYPE is not correctly set."
        )

    if db_type == "elastic-search" and ES_ENABLED:
        # db connection
        elasticlient = ElasticSearchManager(
            host=environ.get("DATABASE_HOST"),
            port=environ.get("DATABASE_PORT"),
            pool_size=environ.get("DATABASE_POOL_CONNECTIONS"),
            user=environ.get("DATABASE_USER"),
            password=environ.get("DATABASE_PASSWORD"),
            ssl=int(environ.get("DATABASE_SSL", "1")),
        )

        # connect
        elasticlient.connect()

        # ensure index is created
        elasticlient.create_index()

        # store db insertion method
        db_insertion = elasticlient.insert_document
    elif db_type == "elastic-search" and not ES_ENABLED:
        exit_cli_error(
            ModuleNotFoundError(
                _("Elastic Search package dependency is not available.")
            )
        )
    else:
        exit_cli_error(
            NotImplementedError(
                _("This database type '{}' is not implemented yet.".format(db_type))
            )
        )

    # -- JSON PROCESSING ---------------------------------------------------------------
    # looking for JSON files
    li_json_files = list(extraction_folder.glob("**/*.json"))
    if not len(li_json_files):
        exit_cli_normal("No JSON files to import")

    # create the output folder
    processed_folder = Path(output_folder / "processed" / "to_{}".format(db_type))
    processed_folder.mkdir(parents=True, exist_ok=True)

    with click.progressbar(
        iterable=li_json_files,
        length=len(li_json_files),
        label="Processing JSON files to {}".format(db_type),
    ) as prgbar_json_to_process:
        for in_json in prgbar_json_to_process:
            # load JSON
            try:
                json_reader = MetadataJsonReader(in_json)
                logger.debug(
                    "Metadata from '{}' found into JSON: {}".format(
                        json_reader.source, json_reader.input_json
                    )
                )
            except Exception as exc:
                logger.error(exc)
                continue

            # import into selected database
            db_insertion(
                document_to_index=json_reader.as_model.to_dict(),
                auto_refresh="wait_for",
                document_type="metadata-lookup",
            )

            # archive loaded JSON
            rename(src=in_json, dst=processed_folder / in_json.name)

        # refresh indices
        elasticlient.es_client.indices.refresh()

    # ending
    exit_cli_success(
        "PROCESS completed after {:5.2f}s.".format(default_timer() - START_TIME)
    )


# #############################################################################
# ##### Stand alone program ########
# ##################################

if __name__ == "__main__":
    """Standalone execution."""
    # additionnal imports
    import multiprocessing

    # workaround for multiprocessing support for packaged version on Windows
    # see: https://github.com/pyinstaller/pyinstaller/wiki/Recipe-Multiprocessing
    # and: https://stackoverflow.com/a/48805137/2556577
    multiprocessing.freeze_support()
    # launch cli
    process(obj={})



#! python3  # noqa: E265

# submodules
from .models import Listing, Lookup, Sign  # noqa: F401
from .json_parser import MetadataJsonReader  # noqa: F401
from .zip_parser import Unzipper  # noqa: F401
